"""Card model package."""
from settings import CARD_CONFIG_FOLDER

from .model import Model


class Card(Model):
    """Card class."""

    yaml_tag = u'!Card'

    def __init__(self, name, nb_input_pins, nb_output_pins):
        """Init card."""
        self.name = name
        self.nb_input_pins = int(nb_input_pins)
        self.nb_output_pins = int(nb_output_pins)

    def save(self):
        """Save card."""
        self.save_yaml(CARD_CONFIG_FOLDER)

    def delete(self):
        """Delete card."""
        self._delete(CARD_CONFIG_FOLDER)

    def __repr__(self):
        """Represent card in order to save it."""
        return "%s(name=%r, nb_input_pins=%r, nb_output_pins=%r)" % (
            self.__class__.__name__, self.name, self.nb_input_pins, self.nb_output_pins)

    @classmethod
    def get_all(cls):
        """Get all card configurations."""
        return cls._get_all(CARD_CONFIG_FOLDER)

    @classmethod
    def get(cls, name):
        """Get card by name."""
        return cls._get(name, CARD_CONFIG_FOLDER)



"""Device model package."""

from serial.tools import list_ports
from settings import DEVICE_CONFIG_FOLDER

from .card import Card
from .model import Model


class Port(Model):
    """Port class."""

    yaml_tag = u'!Port'

    def __init__(self, name, number, enabled, _type="output"):
        """Init port."""
        self.name = name
        self._type = _type
        self.number = number
        self.enabled = enabled if enabled is True or False else (
            True if enabled == 'on' else False)

    def __repr__(self):
        """Represent port in order to save it."""
        return "%s(number=%r, name=%r, _type=%r, enabled=%r)" % (
            self.__class__.__name__, self.number, self.name, self._type, self.enabled)


class Device(Model):
    """Device class."""

    yaml_tag = u'!Device'

    def __init__(self, name, identifier, card_name):
        """Init device."""
        self.name = name
        self.identifier = identifier
        self.card = Card.get(card_name)
        self.init_ports()

    def save(self):
        """Save device."""
        self.save_yaml(DEVICE_CONFIG_FOLDER)

    def delete(self):
        """Delete device."""
        self._delete(DEVICE_CONFIG_FOLDER)

    def __repr__(self):
        """Represent device in order to save it."""
        return "%s(name=%r, identifier=%r, card=%r, ports=%r)" % (
            self.__class__.__name__, self.name, self.identifier, self.card, self.ports)

    def init_ports(self):
        """Init port list."""
        self.ports = {'input': list(), 'output': list()}
        for i in range(self.card.nb_input_pins):
            self.ports['input'].append(Port(
                number=i, name="", enabled="False", _type=""))
        for i in range(self.card.nb_output_pins):
            self.ports['output'].append(Port(
                number=i, name="", enabled="False", _type=""))

    def add_port_from_dict(self, port, dict):
        """Create port and add from dict."""
        port = int(port.replace("port", ""))
        if port >= self.card.nb_input_pins:
            port = port - self.card.nb_input_pins
        p = Port(number=port, **dict)
        _type = p._type if p._type == "output" else "input"
        port_to_replace = next(
            (self.ports[_type].index(port) for port in self.ports[_type] if p.number == port.number), None)
        self.ports[_type][port_to_replace] = p

    def get_filename(self):
        """Get filename to save."""
        return __class__.filenamify(self.identifier) + ".yaml"

    @classmethod
    def get_all(cls):
        """Get all device configurations."""
        return __class__._get_all(DEVICE_CONFIG_FOLDER)

    @classmethod
    def get(cls, name):
        """Get device by name."""
        return __class__._get(name, DEVICE_CONFIG_FOLDER)

    @classmethod
    def get_connected_devices(cls):
        """Get devices connected."""
        serials = list(list_ports.comports())
        arduinos = [s for s in serials if 'Arduino' in s.description]
        devices = {a.device: __class__.get(__class__.get_identifier(
            a)) for a in arduinos}
        return devices

    @classmethod
    def get_config(cls, name):
        """Get config by name."""
        configs = cls.get_all()
        if configs is not None:
            return next(
                (config for config in configs if config.name == name), None)
        else:
            return configs

    @classmethod
    def get_identifier_from_serial(cls, serial):
        """Get device identifier from serial name."""
        serials = list(list_ports.comports())
        arduinos = [s for s in serials if 'Arduino' in s.description]
        arduino = next(
            (arduino for arduino in arduinos if serial in arduino.device), None)
        return __class__.get_identifier(arduino)

    @classmethod
    def get_identifier(cls, arduino):
        """Get identifier from arduino."""
        if arduino is not None:
            config_name = str(arduino.vid)
            config_name += str(arduino.pid)
            config_name += str(arduino.serial_number)
            return config_name
        else:
            return None

    @classmethod
    def from_request_args(cls, slug, args):
        """Init device from request args."""
        args = {arg: args[arg][0].decode() for arg in args}
        res = dict()
        for arg in args:
            if '[' in arg and ']' in arg:
                split_arg = arg.split('[')
                var_name = split_arg[0]
                index = split_arg[1].split(']')[0]
                if var_name not in res:
                    res[var_name] = dict()
                res[var_name][index] = args[arg]
            else:
                res[arg] = args[arg]
        if 'identifier' not in res:
            identifier = cls.get_identifier_from_serial(slug)
        else:
            identifier = res.pop('identifier')
        dev = cls(res.pop('name'), identifier,
                  res.pop('card_name'))
        for port in res:
            dev.add_port_from_dict(port, res[port])
        return dev



"""Index page handler package."""
from tornado.escape import url_unescape

from .baseHandler import BaseHandler


class CrudHandler(BaseHandler):
    """Arduino page handler."""

    def get(self, slug=None, method=None):
        """Handle get request."""
        if slug is None:
            self.list()
            return

        slug = url_unescape(slug)
        if slug == 'create':
            self.create()
        else:
            if method is None:
                self.show(slug)
            elif method == 'edit':
                self.edit(slug)
            elif method == 'create':
                self.create(slug)
            else:
                self.render("404.html")

    def post(self, slug=None, method=None):
        """
        Handle get request.

        Broswers can't send put and delete request so we have to fake them.
        """
        try:
            _method = self.get_argument('_method')
        except Exception:
            _method = None
        if slug is None and _method is None:
            self.store()
        elif _method is None:
            self.store(slug)
        slug = url_unescape(slug)

        if _method == 'put':
            self.put(slug, method)
        elif _method == 'delete':
            self.delete(slug, method)
        else:
            self.render("404.html")

    def put(self, slug, method=None):
        """Handle put request."""
        self.update(slug)

    def delete(self, slug, method=None):
        """Handle delete request."""
        self.destroy(slug)

    def list(self):
        """Show device."""
        pass

    def show(self, slug):
        """Show device."""
        pass

    def create(self, slug=None):
        """Show configuration form for device."""
        pass

    def edit(self, slug):
        """Show configuration form for device."""
        pass

    def store(self, slug):
        """Store configuration."""
        pass

    def update(self, slug):
        """Update configuration."""
        pass

    def destroy(self, slug):
        """Destroy configuration."""
        pass



"""Sensor handler package."""

from models.sensor import Sensor
from tornado.escape import url_escape

from .crudHandler import CrudHandler


class SensorHandler(CrudHandler):
    """Sensor handler."""

    def list(self):
        """List configurations."""
        sensors = Sensor.get_all()
        self.render('sensor/list.html', sensors=sensors)

    def show(self, slug):
        """Show device."""
        sensor = Sensor.get(slug)
        if sensor is not None:
            self.render('sensor/show.html', sensor=sensor, slug=slug)
        else:
            self.redirect('/sensor', permanent=False)

    def create(self):
        """Show configuration form for device."""
        settings = dict()
        settings['sensor'] = None
        settings['method'] = 'post'
        self.render('sensor/config.html', **settings)

    def edit(self, slug):
        """Show configuration form for device."""
        sensor = Sensor.get(slug)
        settings = dict()
        settings['sensor'] = sensor
        settings['method'] = 'put'
        self.render('sensor/config.html', **settings)

    def store(self, slug=""):
        """Create configuration."""
        sensor_name = self.get_argument('name')
        self.save(sensor_name)

    def update(self, slug):
        """Update configuration."""
        self.save(slug)

    def save(self, slug):
        """Save configuration."""
        sensor_name = slug
        min_value = self.get_argument('min_value')
        max_value = self.get_argument('max_value')
        suffix = self.get_argument('suffix')
        reverse = True if 'reverse' in self.request.arguments else False
        sensor = Sensor(sensor_name, min_value, max_value, reverse, suffix)
        sensor.save()
        slug = url_escape(slug)
        redirect_url = self.redirect_url
        if slug not in redirect_url:
            redirect_url += '/' + slug
        self.redirect(redirect_url, permanent=True)

    def destroy(self, slug):
        """Destroy configuration."""
        sensor = Sensor.get(slug)
        sensor.delete()
        self.redirect(self.redirect_url, permanent=False)



"""Arduinozore module."""

import argparse
import os
import socket
import sys
from shutil import copyfile
from shutil import get_terminal_size
from shutil import rmtree

import tornado.httpserver
import tornado.ioloop
import tornado.web
import tornado.websocket
from handlers.error404handler import My404Handler
from handlers.serialManager import SerialManager
from handlers.ws import WSHandler
from pyfiglet import Figlet
from settings import ARDUINO_CODE_FILE_NAME
from settings import ARDUINO_FILE_PATH
from settings import CONFIG_FOLDER
from settings import PORT
from settings import SSL_PORT
from settings import path
from settings import settings
from settings import ssl_opts
from urls import url_pattern


def main():
    """Catch main function."""
    p = argparse.ArgumentParser(
        description="Arduinozore server", prog="arduinozore")
    p.add_argument('-hp',
                   '--http_port',
                   type=int,
                   help='Server http port. Default 80')
    p.add_argument('-hsp',
                   '--https_port',
                   type=int,
                   help='Server https port. Default 443. Used for sockets too.')
    p.add_argument('-a',
                   '--arduino',
                   type=str,
                   metavar='path',
                   help='Path where arduino source code will be generated.')
    p.add_argument('--newconfig',
                   action="store_true",
                   help='Delete actual config and make a new one. Warning.')
    args = p.parse_args()

    if args.arduino:
        copy_arduino_code(args.arduino)

    if args.newconfig:
        if os.path.exists(CONFIG_FOLDER):
            rmtree(CONFIG_FOLDER)

    http_port = PORT if args.http_port is None else args.http_port
    ssl_port = SSL_PORT if args.https_port is None else args.https_port

    check_config_folder()

    serial_manager = SerialManager()
    try:
        if not serial_manager.is_alive():
            serial_manager.start()
        index_application = tornado.web.Application(
            url_pattern, default_handler_class=My404Handler, **settings)

        index_application.listen(http_port)
        http_server = tornado.httpserver.HTTPServer(
            index_application,
            ssl_options=ssl_opts
        )
        http_server.listen(ssl_port)
        tornado.ioloop.PeriodicCallback(WSHandler.write_to_clients, 500).start()

        terminal_width = get_terminal_size((80, 20))[0]
        introduction(ssl_port, terminal_width)

        tornado.ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        print('Exiting...'.center(terminal_width))
    except Exception as e:
        print(e)
        sys.exit()
    finally:
        serial_manager.join()


def introduction(ssl_port, terminal_width):
    """Show a message to the user so he knows who we are."""
    app_name = '  Arduinozore'
    TOPBAR = '#' * (terminal_width - 2)

    print('/' + TOPBAR + '\\' + "\n")
    print(Figlet(font='banner').renderText(app_name))
    print('\\' + TOPBAR + '/' + "\n")
    print('/' + TOPBAR + '\\' + "\n")
    HOST = socket.gethostname()
    PORT = ssl_port
    message = 'Listening on: https://{}:{}'.format(HOST, PORT)
    print(message.center(terminal_width))
    sys.stdout.flush()


def check_config_folder():
    """Check if config folder exists, otherwise creates it."""
    try:
        if not os.path.exists(CONFIG_FOLDER):
            os.makedirs(CONFIG_FOLDER)
    except Exception as e:
        exit(e)


def copy_arduino_code(dest):
    """Copy arduino source code to dest."""
    dst = path(dest, ARDUINO_CODE_FILE_NAME)
    copyfile(ARDUINO_FILE_PATH, dst)
    print("File copied to " + dst)
    exit(0)


if __name__ == "__main__":
    main()



"""Urls module."""
from handlers.card import CardHandler
from handlers.device import DevicePageHandler
from handlers.index import IndexPageHandler
from handlers.sensor import SensorHandler
from handlers.setting import SettingPageHandler
from handlers.ws import WSHandler
from settings import STATIC_PATH
from tornado.web import StaticFileHandler

url_pattern = [
    (r'/static/(.*)', StaticFileHandler, {'path': STATIC_PATH}),
    (r'/', IndexPageHandler),
    (r'/settings/?$', SettingPageHandler),
    (r'/device/?$', DevicePageHandler),
    (r'/device/((?!create|/)[^/]+)/?$', DevicePageHandler),
    (r'/device/([^/]+)/(?:(edit|create)\w?)?/?', DevicePageHandler),
    (r'/sensor/?$', SensorHandler),
    (r'/sensor/([^/]+)/?$', SensorHandler),
    (r'/sensor/([^/]+)/(?:(edit)\w?)?/?$', SensorHandler),
    (r'/card/?$', CardHandler),
    (r'/card/([^/]+)/?$', CardHandler),
    (r'/card/([^/]+)/(?:(edit)\w?)?/?$', CardHandler),
    (r'/ws/([^/]+)', WSHandler),
]



"""Index page handler package."""
from settings import TEMPLATE_FOLDER
from tornado.web import RequestHandler


class BaseHandler(RequestHandler):
    """Device page handler."""

    def get_template_path(self):
        """Get template path in order to find template files."""
        return TEMPLATE_FOLDER

    def prepare(self):
        """Prepare requests."""
        self.redirect_url = 'https://' + self.request.host
        self.redirect_url += self.request.uri.rstrip("/")

        if self.request.protocol == 'http':
            self.redirect(self.redirect_url, permanent=False)

    def set_default_headers(self, *args, **kwargs):
        """Set default headers."""
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header("Access-Control-Allow-Methods", "GET, OPTIONS")
        self.set_header("Strict-Transport-Security",
                        "max-age=500; includeSubDomains; preload")



"""Logging process."""
import json
import sys
import time
from multiprocessing import Event
from multiprocessing import Manager
from multiprocessing import Process

from .serialReader import SerialReader


class Singleton(type):
    """
    Singleton metaclass.

    From https://stackoverflow.com/q/6760685/9395299.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Instantiate singleton or return."""
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SerialManager(Process, metaclass=Singleton):
    """Process class."""

    def __init__(self):
        """Init process."""
        Process.__init__(self)
        self.exit = Event()
        self.serial_readers = {}
        self.datas = Manager().dict()
        self.out = Manager().dict()

    def toggle_pin(self, port, pin):
        """Toggle pin on card connected to port."""
        if port not in dict(self.out):
            self.out[port] = pin

        print(port)
        print(self.out[port])
        sys.stdout.flush()

    def get_toggelable_pin(self, port):
        """Return toggelable pins for port."""
        try:
            toggle = self.out[port]
            del self.out[port]
            return toggle
        except KeyError:
            return None

    def set_datas(self, port, datas):
        """Set data from child process."""
        self.datas[port] = json.dumps(datas)

    def get_serial_reader(self, port):
        """Get datas for specified serial port."""
        if port not in self.serial_readers:
            self.serial_readers[port] = SerialReader(port, self)
            self.serial_readers[port].start()

    def get_datas_for_port(self, port):
        """Get datas for serial port."""
        self.get_serial_reader(port)
        try:
            return self.datas[port]
        except KeyError:
            self.datas[port] = 'Initializing reader'
            return self.datas[port]

    def run(self):
        """Run process."""
        while not self.exit.is_set():
            try:
                # print("Manager running")
                sys.stdout.flush()
                time.sleep(1)
            except (KeyboardInterrupt, RuntimeError) as e:
                self.shutdown()
            except Exception as e:
                raise e
            finally:
                for s_r in self.serial_readers:
                    self.serial_readers[s_r].join()

    def shutdown(self):
        """Shut down process."""
        self.exit.set()



"""Index page handler package."""

from models.device import Device

from .baseHandler import BaseHandler


class IndexPageHandler(BaseHandler):
    """Index page handler."""

    def get(self):
        """Handle get request."""
        devices = Device.get_connected_devices()
        self.render('index.html', devices=devices)



"""Index page handler package."""
import os
import socket

from models.card import Card
from models.device import Device
from models.sensor import Sensor
from settings import DEVICE_CONFIG_FOLDER
from settings import SSL_PORT
from settings import path

from .crudHandler import CrudHandler
from .tools import get_arduino
from .tools import get_config_name


class DevicePageHandler(CrudHandler):
    """Device page handler."""

    default_args = {'enabled': '', 'name': '', 'type': ''}

    def list(self):
        """List configuration."""
        devices = Device.get_all()
        self.render('device/list.html', devices=devices)

    def show(self, slug):
        """Show device."""
        device = Device.get(Device.get_identifier_from_serial(slug))
        if device is None:
            device = Device.get_config(slug)
            if device is None:
                self.redirect(self.redirect_url + '/create', permanent=False)
            else:
                settings = dict()
                settings['device'] = device
                settings['slug'] = slug
                self.render('device/show.html', **settings)
        else:
            the_host = socket.gethostname()
            settings = dict()
            settings['host'] = the_host
            settings['port'] = SSL_PORT
            settings['slug'] = slug
            settings['device'] = device
            self.render('device/communicate.html', **settings)

    def create(self, slug):
        """Show configuration form for device."""
        cards = Card.get_all()
        sensors = Sensor.get_all()
        device = Device.get(slug)
        if 'card' in self.request.arguments:
            card = Card.get(self.get_argument('card'))
        else:
            card = None
        settings = dict()
        settings['method'] = 'post'
        settings['cards'] = cards
        settings['card'] = card
        settings['sensors'] = sensors
        settings['device'] = device
        settings['slug'] = slug
        settings['method'] = 'post'
        self.render('device/config.html', **settings)

    def edit(self, slug):
        """Show configuration form for device."""
        device = Device.get(Device.get_identifier_from_serial(slug))
        cards = Card.get_all()
        sensors = Sensor.get_all()
        if device is None:
            device = Device.get_config(slug)
            if device is None:
                self.redirect(self.redirect_url + '/create', permanent=False)

        settings = dict()
        settings['method'] = 'put'
        settings['cards'] = cards
        settings['card'] = device.card
        settings['sensors'] = sensors
        settings['device'] = device
        settings['method'] = 'put'
        self.render('device/config.html', **settings)

    def store(self, slug):
        """Store configuration."""
        self.save(slug)

        self.redirect(self.redirect_url, permanent=True)

    def update(self, slug):
        """Update configuration."""
        self.save(slug)

        self.redirect(self.redirect_url, permanent=True)

    def save(self, slug):
        """Save configuration."""
        try:
            self.request.arguments.pop("_method")
        except Exception:
            pass
        device = Device.from_request_args(slug, self.request.arguments)
        device.save()

    def destroy(self, slug):
        """Destroy configuration."""
        arduino = get_arduino(slug)
        config_name = get_config_name(arduino)
        config_file = path(DEVICE_CONFIG_FOLDER, config_name)
        os.remove(config_file)
        self.redirect(self.redirect_url, permanent=False)



"""Arduinozore setup file."""

from subprocess import check_call

from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install

INSTALL_CMD = "sudo apt-get install openssl"
CERT_CMD = "./install_cert.sh"


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        """Run post install script."""
        check_call(INSTALL_CMD.split())
        check_call(CERT_CMD.split())
        develop.run(self)


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        """Run post install script."""
        check_call(INSTALL_CMD.split())
        check_call(CERT_CMD.split())
        install.run(self)


setup(
    setup_requires=['pbr'],
    pbr=True,
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    },
)



"""Sensor model package."""
from settings import SENSOR_CONFIG_FOLDER

from .model import Model


class Sensor(Model):
    """Sensor class."""

    yaml_tag = u'!Sensor'

    def __init__(self, name, min_value, max_value, reverse=False, suffix=""):
        """Init sensor."""
        self.name = name
        self.min_value = int(min_value)
        self.max_value = int(max_value)
        self.reverse = reverse
        self.suffix = suffix

    def save(self):
        """Save sensor."""
        self.save_yaml(SENSOR_CONFIG_FOLDER)

    def delete(self):
        """Delete sensor."""
        self._delete(SENSOR_CONFIG_FOLDER)

    def __repr__(self):
        """Represent sensor in order to save it."""
        return "%s(name=%r, min_value=%r, max_value=%r, reverse=%r, suffix=%r)" % (
            self.__class__.__name__, self.name, self.min_value, self.max_value,
            self.reverse, self.suffix)

    def transform_datas(self, raw_datas):
        """Transform raw datas."""
        d = float(raw_datas)
        if self.reverse:
            return str(round((d / 1024 * (self.min_value-self.max_value) + self.max_value), 1)) + " " + self.suffix
        else:
            return str(round((d / 1024 * (self.max_value-self.min_value) + self.min_value), 1)) + " " + self.suffix

    @classmethod
    def get_all(cls):
        """Get all sensor configurations."""
        return __class__._get_all(SENSOR_CONFIG_FOLDER)

    @classmethod
    def get(cls, name):
        """Get sensor by name."""
        return __class__._get(name, SENSOR_CONFIG_FOLDER)
